from __future__ import annotations

from jsonscreamer.compile import _name_from_validator


def test_name_from_validator() -> None:
    def not_():
        pass

    def a_long_name():
        pass

    def x_123_a():
        pass

    assert _name_from_validator(not_) == "not"
    assert _name_from_validator(a_long_name) == "aLongName"
    assert _name_from_validator(x_123_a) == "x123A"



from __future__ import annotations

import json as _json
import pathlib as _pathlib
from typing import TYPE_CHECKING as _TYPE_CHECKING

from . import array, basic, compile, logical, object_
from .format import FORMATS as _FORMATS
from .resolve import HANDLERS as _HANDLERS, RefTracker as _RefTracker
from .types import Context as _Context

if _TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from typing import Any

    from .types import Format, Json, Schema, ValidationError

    Handler = Callable[[str], Json]


class Validator:
    """Validates instances against a given schema.

    Usage:
        >>> validator = Validator(some_schema)
        >>> assert validator.is_valid(some_instance)
        >>> validator.validate(some_instance)
    """

    _metavalidator = None

    def __init__(
        self,
        schema: Schema | bool = True,
        formats: dict[str, Format] | bool = True,
        handlers: dict[str, Handler] | None = None,
        check_schema: bool = True,
    ) -> None:
        if check_schema:
            type(self).check_schema(schema)

        if formats is False:
            formats = {}  # completely disable format checking
        elif formats is True:
            formats = _FORMATS  # use default format checking
        else:
            formats = _FORMATS | formats  # extend with custom formats

        handlers = _HANDLERS | (handlers or {})
        tracker = _RefTracker(schema, handlers=handlers)
        self._context = _Context(formats=formats, tracker=tracker)

        # NOTE: if there were no $ref item in the schema, we wouldn't need a tracker,
        # it just obscures the logic. However, given that refs exist and can be circular
        # we have to track where we are and where we've been within the schemas:
        while tracker:
            uri = tracker.pop()
            with tracker._resolver.resolving(uri) as sub_defn:
                tracker.compiled[uri] = compile.compile_(sub_defn, self._context)

        self._validator = tracker.entrypoint

    # Simple validation functions:

    def is_valid(self, instance: Any) -> bool:
        """Check whether the given instance is valid."""
        return not any(self._validator(instance, []))

    def validate(self, instance: Any) -> None:
        """Validate the instance and raise a ValidationError if it is invalid."""
        for err in self._validator(instance, []):
            raise err

    def iter_errors(self, instance: Any) -> Iterator[ValidationError]:
        """Iterate over all validation errors for the instance."""
        yield from self._validator(instance, [])

    # These are a little more baroque - but basically aimed at loading / creating
    # a schema validator at most once:

    @classmethod
    def check_schema(cls, schema: Any) -> None:
        """Validate the schema and raise a ValidationError if it is invalid.

        Handles only the draft 07 spec.
        """
        return cls.metavalidator().validate(schema)

    @classmethod
    def metavalidator(cls) -> Validator:
        """Return the Validator for draft 07 schemas."""
        if cls._metavalidator is None:
            here = _pathlib.Path(__file__).parent
            with (here / "_metaschemas" / "draft07.json").open() as f:
                metaschema = _json.load(f)

            cls._metavalidator = cls(metaschema, check_schema=False)

        return cls._metavalidator


__all__ = ["Validator", "array", "basic", "compile", "logical", "object_"]



from __future__ import annotations

import pathlib
from http.server import HTTPServer, SimpleHTTPRequestHandler
from multiprocessing import Process

import pytest

HERE = pathlib.Path(__file__).parent.resolve()
REMOTES = HERE / "suite" / "JSON-Schema-Test-Suite" / "remotes"


class _RemotesRequestHandler(SimpleHTTPRequestHandler):
    """Uses remote schemas directory as the default."""

    def __init__(self, request, client_address, server, *, directory=REMOTES):
        super().__init__(request, client_address, server, directory=directory)


def _remote_ref_server_impl():
    httpd = HTTPServer(("localhost", 1234), _RemotesRequestHandler)
    httpd.serve_forever()


@pytest.fixture(scope="session")
def _remote_ref_server():
    """Runs a server on localhost to serve "remote" schemas."""
    proc = Process(target=_remote_ref_server_impl, daemon=True)
    proc.start()



from __future__ import annotations

from jsonscreamer import Validator
from jsonscreamer.compile import compile_
from jsonscreamer.resolve import RefTracker
from jsonscreamer.types import Context

POST_BODY = {
    "id": 0,
    "category": {"id": 0, "name": "string"},
    "name": "doggie",
    "photoUrls": ["string"],
    "tags": [{"id": 0, "name": "string"}],
    "status": "available",
}


SCHEMA = {
    "type": "object",
    "required": ["name", "photoUrls"],
    "properties": {
        "id": {
            "type": "integer",
        },
        "name": {
            "type": "string",
        },
        "category": {
            "type": "object",
            "properties": {
                "id": {"type": "integer"},
                "name": {"type": "string"},
            },
            "required": ["id", "name"],
        },
        "photoUrls": {
            "type": "array",
            "items": {"type": "string"},
        },
        "tags": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "id": {"type": "integer"},
                    "name": {"type": "string"},
                },
                "required": ["id", "name"],
            },
        },
        "status": {
            "type": "string",
            "enum": [
                "available",
                "pending",
                "sold",
            ],
        },
    },
}


def test_complex():
    validator = compile_(SCHEMA, Context({}, RefTracker(SCHEMA, {})))

    assert list(validator("fish", []))
    assert list(validator({}, []))
    assert list(validator(POST_BODY, [])) == []
    assert list(validator({**POST_BODY, "name": 3}, []))
    assert list(validator({**POST_BODY, "category": {"name": "fish"}}, []))


def test_iter_errors():
    bad_instance = {
        "id": "bad",  # error 1
        "name": 42,  # error 2
        "category": {"id": "nested-bad"},  # errors 3 & 4
        "photoUrls": [1, 2],  # errors 5 and 6
        "tags": [{}, {"id": "spam", "name": "spam"}],  # errors 7, 8, 9
        "status": "sold",
    }
    validator = Validator(SCHEMA)

    errors = list(validator.iter_errors(bad_instance))
    assert len(errors) == 9

    actual_errors = [(e.absolute_path, e.type) for e in errors]
    expected_errors = [
        (("id",), "type"),
        (("name",), "type"),
        (("category", "id"), "type"),
        (("category",), "required"),
        (("photoUrls", 0), "type"),
        (("photoUrls", 1), "type"),
        (("tags", 0), "required"),
        (("tags", 0), "required"),
        (("tags", 1, "id"), "type"),
    ]

    for actual, expected in zip(actual_errors, expected_errors):
        assert actual == expected



"""Functions to build validators for basic types of the JSON schema.

For example `enum({"type": "string", "enum": ["foo", "bar"]})`
should return a validator function like this:
```
def validate(value):
    return value in {"foo", "bar"}
```
"""

from __future__ import annotations

import functools as _functools
import logging as _logging
import re as _re
from typing import TYPE_CHECKING as _TYPE_CHECKING

from .compile import register as _register
from .types import ValidationError

if _TYPE_CHECKING:
    from collections.abc import Callable, Collection, Iterable
    from typing import Any

    from .types import Context, Json, Path, Schema, Validator


class _StrictBool:
    """Exists purely because 0 == False in python.

    Casting to this type will ensure that 0 != False, [0] != [False] etc.
    """

    def __init__(self, value: float | bool) -> None:
        self.value = value

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self) -> str:
        return f"<_StrictBool({self.value!r})>"

    def __hash__(self) -> int:
        return hash(self.value)

    def __eq__(self, other: object) -> bool:
        return (
            self.value is other
            or (
                self.value == other
                and not isinstance(self.value, bool)
                and not isinstance(other, bool)
            )
            or (
                isinstance(other, self.__class__)
                and (
                    self.value is other.value
                    or (
                        self.value == other.value
                        and not isinstance(self.value, bool)
                        and not isinstance(other.value, bool)
                    )
                )
            )
        )

    __req__ = __eq__


__true = _StrictBool(True)
__false = _StrictBool(False)
__one = _StrictBool(1)
__zero = _StrictBool(0)


def _strict_bool_nested(x: object) -> object:
    """Convert possible bools in nested data structures to _StrictBool"""
    if isinstance(x, list):
        return [_strict_bool_nested(v) for v in x]
    elif isinstance(x, dict):
        return {k: _strict_bool_nested(v) for k, v in x.items()}
    elif x is True:
        return __true
    elif x is False:
        return __false
    elif x == 1:
        return __one
    elif x == 0:
        return __zero
    return x


def _min_len_validator(n: int, kind: str) -> Validator:
    def validate(x: Json, path: Path) -> Iterable[ValidationError]:
        if len(x) < n:  # type: ignore (assumption: sized object provided)
            yield ValidationError(
                tuple(path), f"{x} is too short (min length {n})", kind
            )

    return validate


def _max_len_validator(n: int, kind: str) -> Validator:
    def validate(x: Json, path: Path) -> Iterable[ValidationError]:
        if len(x) > n:  # type: ignore (assumption: sized object provided)
            yield ValidationError(
                tuple(path), f"{x} is too long (max length {n})", kind
            )

    return validate


@_register
def min_length(defn: Schema, context: Context) -> Validator | None:
    value: int = defn["minLength"]
    guard = _string_guard(defn)
    return guard(_min_len_validator(value, "minLength"))


@_register
def max_length(defn: Schema, context: Context) -> Validator | None:
    value: int = defn["maxLength"]
    guard = _string_guard(defn)
    return guard(_max_len_validator(value, "maxLength"))


@_register
def pattern(defn: Schema, context: Context) -> Validator | None:
    value: str = defn["pattern"]
    rex = _re.compile(value)

    @_string_guard(defn)
    def validate(x: str, path: Path) -> Iterable[ValidationError]:
        if not rex.search(x):
            yield ValidationError(
                tuple(path), f"'{x}' does not match pattern '{value}'", "pattern"
            )

    return validate


@_register
def enum(defn: Schema, context: Context) -> Validator:
    value: list[object] = defn["enum"]

    members: Collection[object] = list(map(_strict_bool_nested, value))
    try:
        members = set(members)
    except TypeError:
        pass  # Unhashable have to use O(n) lookup

    def validate(x: Json, path: Path) -> Iterable[ValidationError]:
        try:
            if x not in members:
                yield ValidationError(
                    tuple(path), f"'{x}' is not one of {value}", "enum"
                )
        except TypeError:  # checking if unhashable type in a set of hashable objects
            yield ValidationError(tuple(path), f"'{x}' is not one of {value}", "enum")

    return validate


@_register
def const(defn: Schema, context: Context) -> Validator:
    value: object = _strict_bool_nested(defn["const"])

    def validate(x: Json, path: Path) -> Iterable[ValidationError]:
        if x != value:
            yield ValidationError(tuple(path), f"{x} is not {value}", "const")

    return validate


@_register
def format_(defn: Schema, context: Context) -> Validator | None:
    value: str = defn["format"]
    if value in context.formats:
        format = context.formats[value]

        @_string_guard(defn)
        def validate(x: str, path: Path) -> Iterable[ValidationError]:
            if not format(x):
                yield ValidationError(
                    tuple(path), f"{x} does not match format '{value}", "format"
                )

        return validate

    _logging.warning(f"Unsupported format ({value}) will not be checked")


@_register
def minimum(defn: Schema, context: Context) -> Validator | None:
    value: float | int = defn["minimum"]

    @_number_guard(defn)
    def validate(x: float, path: Path) -> Iterable[ValidationError]:
        if x < value:
            yield ValidationError(tuple(path), f"{x} < {value}", "minimum")

    return validate


@_register
def exclusive_minimum(defn: Schema, context: Context) -> Validator | None:
    value: float | int = defn["exclusiveMinimum"]

    @_number_guard(defn)
    def validate(x: float, path: Path) -> Iterable[ValidationError]:
        if x <= value:
            yield ValidationError(tuple(path), f"{x} <= {value}", "exclusiveMinimum")

    return validate


@_register
def maximum(defn: Schema, context: Context) -> Validator | None:
    value: float | int = defn["maximum"]

    @_number_guard(defn)
    def validate(x: float, path: Path) -> Iterable[ValidationError]:
        if x > value:
            yield ValidationError(tuple(path), f"{x} > {value}", "maximum")

    return validate


@_register
def exclusive_maximum(defn: Schema, context: Context) -> Validator | None:
    value: float | int = defn["exclusiveMaximum"]

    @_number_guard(defn)
    def validate(x: float, path: Path) -> Iterable[ValidationError]:
        if x >= value:
            yield ValidationError(tuple(path), f"{x} >= {value}", "exclusiveMaximum")

    return validate


@_register
def multiple_of(defn: Schema, context: Context) -> Validator | None:
    value: float | int = defn["multipleOf"]

    @_number_guard(defn)
    def validate(x: float, path: Path) -> Iterable[ValidationError]:
        # More accurate than x % multiplier == 0
        try:
            frac = x / value
            if int(frac) == frac:
                return None
        except OverflowError:
            pass

        yield ValidationError(
            tuple(path), f"{x} is not a multiple of {value}", "multipleOf"
        )

    return validate


def _type_guard(
    defn: Schema, schema_types: tuple[str, ...], py_types: tuple[type, ...] | type
) -> ...:
    """Create a type guard suitable for decorating a validator.

    We are lazy in our type checking: failing a type check at the schema level
    will always short-circuit the validations and stop further checks. This means
    we can be confident these validators will always be passed e.g. str even without
    performing the type check directly.

    Usage:
        @_type_guard(defn, schema_types=("integer",), py_types=int)
        def validator(...)
            ...
    """

    def decorator(
        validator: Callable[[Any, Path], Iterable[ValidationError]],
    ) -> Validator:
        if "type" in defn:
            types = (
                {defn["type"]} if isinstance(defn["type"], str) else set(defn["type"])
            )
            if types.issubset(schema_types):
                # we can omit type check as it's guaranteed by the schema
                return validator  # pyright: ignore[reportReturnType] (protected by schema type)

        @_functools.wraps(validator)
        def guarded(x: Json, path: Path) -> Iterable[ValidationError]:
            if isinstance(x, py_types):
                yield from validator(x, path)

        return guarded

    return decorator


_number_guard = _functools.partial(
    _type_guard, schema_types=("number", "integer"), py_types=(int, float)
)
_string_guard = _functools.partial(_type_guard, schema_types=("string",), py_types=str)



from __future__ import annotations

from jsonscreamer import Validator


def test_contains():
    schema = {"contains": {"type": "integer"}}
    validator = Validator(schema)

    assert not validator.is_valid([{}])
    assert not validator.is_valid([{}, "1"])
    assert validator.is_valid([{}, False, 1])


def test_max_contains():
    schema = {
        "contains": {"type": "integer"},
        "maxContains": 2,
    }
    validator = Validator(schema)

    assert not validator.is_valid([[], False])
    assert validator.is_valid([1])
    assert validator.is_valid([1, 2])
    assert not validator.is_valid([1, 2, 3])



from __future__ import annotations

import json
import os
import pathlib

import pytest

from jsonscreamer import Validator

HERE = pathlib.Path(__file__).parent
TEST_SUITE = HERE / "JSON-Schema-Test-Suite" / "tests"
DRAFT7 = TEST_SUITE / "draft7"
TEST_FILES = tuple(fn for fn in os.listdir(DRAFT7) if fn.endswith(".json"))
XFAILS = frozenset(
    (
        "refs with relative uris and defs",
        "relative refs with absolute uris and defs",
    )
)


def _enumerate_test_cases():
    for filename in TEST_FILES:
        with (DRAFT7 / filename).open() as f:
            test_cases = json.load(f)

        for test_case in test_cases:
            yield pytest.param(test_case, id=test_case["description"])


@pytest.mark.usefixtures("_remote_ref_server")
@pytest.mark.parametrize("test_case", _enumerate_test_cases())
def test_draft7(test_case):
    if (description := test_case["description"]) in XFAILS:
        pytest.xfail(f"{description} is not yet supported")

    schema = test_case["schema"]
    validator = Validator(schema)

    for test in test_case["tests"]:
        if test["valid"]:
            assert validator.is_valid(test["data"]), test["description"]
        else:
            assert not validator.is_valid(test["data"]), test["description"]



from __future__ import annotations

from unittest import mock

from jsonscreamer.format import FORMATS
from jsonscreamer.logical import all_of, any_of, not_, one_of


def test_not():
    assert list(not_({"not": False}, mock.Mock())("spam", [])) == []
    assert list(not_({"not": True}, mock.Mock())("spam", []))


def test_all_of():
    validator = all_of(
        {
            "allOf": [
                {"type": "string"},
                {"format": "email"},
                True,
            ],
        },
        mock.Mock(formats=FORMATS),
    )

    assert list(validator("alice@bob.com", [])) == []
    assert list(validator("alice", []))
    assert list(validator(42, []))


def test_any_of():
    validator = any_of(
        {
            "anyOf": [
                {"type": "string"},
                {"type": "integer"},
                False,
            ]
        },
        mock.Mock(),
    )

    assert list(validator("42", [])) == []
    assert list(validator(42, [])) == []
    assert list(validator(True, []))


def test_one_of():
    validator = one_of(
        {
            "oneOf": [
                {"required": ["spam"]},
                {"required": ["eggs"]},
            ]
        },
        mock.Mock(),
    )

    assert list(validator({"spam": 42}, [])) == []
    assert list(validator({"eggs": 42}, [])) == []
    assert list(validator({"spam": 42, "eggs": 42}, []))
    assert list(validator({}, []))



from __future__ import annotations

import sys

import pytest

from jsonscreamer import format

_TEST_CASES = {
    "is_date_time": [
        ("2020-01-02T03:04:05+00:00", True),
        ("2020-01-02T03:04:05Z", True),
        ("2020-02-29T03:04:05-03:00", True),
        ("2021-02-29T03:04:05-03:00", False),
        ("2020-01-01 01:01:01Z", False),
        ("2020-01-02", False),
        ("blorp", False),
    ],
    "is_date": [
        ("2020-01-02", True),
        ("2020-01-02Z", False),
        ("2020-02-29", True),
        ("2021-02-29", False),
        ("2020-01", False),
        ("blorp", False),
    ],
    "is_time": [
        ("03:04:05+00:00", True),
        ("03:04:05Z", True),
        ("03:04:05-03:00", True),
        ("T03:04:05-03:00", False),
        (" 01:01:01Z", False),
        ("2020-01-02T03:04:05Z", False),
        ("blorp", False),
    ],
    "is_uuid": [
        ("6e6659ec-4503-4428-9f03-2e2ea4d6c278", True),
        ("6e6659ez-4503-4428-9f03-2e2ea4d6c278", False),
        ("6e6659ec450344289f032e2ea4d6c278", False),
        ("blorp", False),
    ],
    "is_email": [
        ("bob@something.org", True),
        ("bob", False),
    ],
    "is_hostname": [
        ("bla.net", True),
        ("google.com", True),
        ("ドメイン.テスト", False),
        ("", False),
        ("a*", False),
        ("blorp", False),  # FIXME: see below
        ("🐵.com", False),
    ],
    "is_idn_host_name": [
        ("bla.net", True),
        ("google.com", True),
        ("ドメイン.テスト", True),
        ("", False),
        ("a*", False),
        ("blorp", True),  # FIXME: this differs from fqdn - one must be wrong?
        ("🐵.com", False),
    ],
    "is_uri_template": [
        ("https://bla.bla.com/user", True),
        ("https://bla.bla.com/user{/login}", True),
        ("https://bla.bla.com/user{/login", False),
        ("https://bla.bla.com/user{}", False),
    ],
    "is_duration": [
        ("P1000Y12M08DT55H13M56S", True),
        ("P1000Y12M08DT55H13M56.2S", True),
        ("P1000Y12M08DT55H13M56,2S", True),
        ("P10Y10DT10M", True),
        ("P1000Y", True),
        ("PT10M", True),
        ("PT0.5M", True),
        ("P0,5M", True),
        ("P10D10M", False),
        ("PT2S2H", False),
        ("P0,5YT0,5H", False),
        ("P1,000Y0,5M", False),
        ("1Y", False),
        ("P", False),
    ],
    "is_ipv4": [
        ("blorp", False),
        ("10.10.10.10", True),
    ],
    "is_ipv6": [
        ("blorp", False),
        ("10.10.10.10", False),
        ("2001:db8::1000", True),
    ],
    "is_regex": [
        ("(", False),
        ("blorp", True),
        (r"\((.+)", True),
    ],
    "is_uri": [
        ("http://xn--rsum-bpad.example.org", True),
        ("http://ドメイン.テスト", False),
        ("//lol", False),
    ],
    "is_uri_reference": [
        ("//bla", True),
        ("ha ha", False),
        ("//テスト", False),
    ],
    "is_iri": [
        ("http://xn--rsum-bpad.example.org", True),
        ("http://ドメイン.テスト", True),
        ("lol", False),
    ],
    "is_iri_reference": [
        ("//bla", True),
        ("//テスト", True),
        ("ha ha", False),
    ],
    "is_json_pointer": [
        ("/bla/0", True),
        ("/", True),
        ("", True),
        ("1/bob", False),
        ("bob", False),
    ],
    "is_relative_json_pointer": [
        ("1/bla/0", True),
        ("2/", True),
        ("3#", True),  # FIXME: I don't think this is allowed
        ("4", True),
        ("0/", True),
        ("01/", False),
        ("-2/", False),
        ("bob", False),
    ],
}


def get_test_cases():
    for func_name, test_cases in _TEST_CASES.items():
        for value, valid in test_cases:
            yield func_name, value, valid


@pytest.mark.parametrize("func_name,value,valid", get_test_cases())
def test_format(func_name, value, valid):
    func = getattr(format, func_name)
    if valid:
        assert func(value)
    else:
        assert not func(value)


def test_format_lookup():
    assert format.FORMATS["date-time"] == format.is_date_time
    assert format.FORMATS["uuid"] == format.is_uuid


def test_tests():
    # A meta-test to test we're testing all of the formats
    for func in format.FORMATS.values():
        assert _TEST_CASES.get(func.__name__), (
            f"You need to write a test for {func.__name__}"
        )


@pytest.mark.parametrize(
    "value,valid",
    (
        ("2020", False),
        ("2020-01", False),
        ("2020-01-02", True),
        ("2020-01-02T01", True),
        ("2020-01-02T01:02:03", True),
        ("2020-01-02T01:02:03.456", True),
        ("2020-01-02T01:02:03.456Z", sys.version_info >= (3, 11)),  # 3.10 bug
        ("2020-01-02T01:02:03.456+00:01", True),
        ("2020-01-02T01:02:66.456+00:01", False),
        ("2020-01-02 01", True),
        ("2020-01-02 01:02:03", True),
        ("2020-01-02 01:02:03.456", True),
        ("2020-01-02 01:02:03.456+00:01", True),
        ("2020-01-02 01:02:66.456+00:01", False),
    ),
)
def test_is_date_time_iso(value, valid):
    assert format.is_date_time_iso(value) == valid



from __future__ import annotations

import datetime
from unittest import mock

import pytest

from jsonscreamer import Validator
from jsonscreamer.basic import (
    _StrictBool,
    const,
    enum,
    exclusive_maximum,
    exclusive_minimum,
    format_,
    max_length,
    maximum,
    min_length,
    minimum,
    multiple_of,
    pattern,
)
from jsonscreamer.compile import type_
from jsonscreamer.format import FORMATS

TYPENAMES = ("boolean", "integer", "null", "number", "string", "zzzzzz")
_VALID = {
    "boolean": (bool,),
    "integer": (int,),
    "null": (type(None),),
    "number": (float, int),
    "string": (str,),
    "array": (list,),
    "object": (dict,),
}


@pytest.mark.parametrize("typename", _VALID)
def test_validate_type(typename):
    validator = type_({"type": typename}, mock.Mock())
    valid = _VALID[typename]

    for value in [
        -100,
        0,
        100,
        None,
        -1.1,
        3.14,
        "lol",
        "",
        datetime.datetime(2018, 1, 1),
        {},
        [],
    ]:
        result = validator(value, list("xy"))  # pyright: ignore[reportArgumentType]

        if isinstance(value, valid):
            assert not result, f"{value} should be a valid {typename}"
        else:
            assert result, f"{value} should not be a valid {typename}"

    for value in (True, False):
        result = validator(value, list("xy"))
        if typename == "boolean":
            assert not result
        else:
            assert result


def test_min_max_length():
    defn = {"type": "string", "minLength": 3}
    validator = min_length(defn, mock.Mock())
    assert validator

    assert list(validator("fo", []))
    assert list(validator("foo", [])) == []
    assert list(validator("fooooo", [])) == []

    defn = {"type": "string", "maxLength": 3}
    validator = max_length(defn, mock.Mock())
    assert validator

    assert list(validator("fo", [])) == []
    assert list(validator("foo", [])) == []
    assert list(validator("fooooo", []))


def test_pattern():
    defn = {
        "type": "string",
        "minLength": 2,
        "maxLength": 20,
        "pattern": r"^([a-z]+)@([a-z]+)\.com$",
    }
    validator = pattern(defn, mock.Mock())
    assert validator

    assert list(validator("", []))
    assert list(validator("foo@bar.com", [])) == []
    assert list(validator(" foo@bar.com", []))
    assert list(validator("foo@bar.com etc", []))


def test_min():
    defn = {"type": "number", "minimum": 3}
    validator = minimum(defn, mock.Mock())
    assert validator
    assert list(validator(0, []))
    assert list(validator(3, [])) == []
    assert list(validator(5, [])) == []
    assert list(validator(7, [])) == []
    assert list(validator(10, [])) == []

    defn = {"type": "number", "exclusiveMinimum": 3}
    validator = exclusive_minimum(defn, mock.Mock())
    assert validator
    assert list(validator(0, []))
    assert list(validator(3, []))
    assert list(validator(5, [])) == []
    assert list(validator(7, [])) == []
    assert list(validator(10, [])) == []


def test_max():
    defn = {"type": "number", "maximum": 7}
    validator = maximum(defn, mock.Mock())
    assert validator
    assert list(validator(0, [])) == []
    assert list(validator(3, [])) == []
    assert list(validator(5, [])) == []
    assert list(validator(7, [])) == []
    assert list(validator(10, []))

    defn = {"type": "number", "exclusiveMaximum": 7}
    validator = exclusive_maximum(defn, mock.Mock())
    assert validator
    assert list(validator(0, [])) == []
    assert list(validator(3, [])) == []
    assert list(validator(5, [])) == []
    assert list(validator(7, []))
    assert list(validator(10, []))


def test_multiple():
    defn = {"type": "number", "multipleOf": 3}
    validator = multiple_of(defn, mock.Mock())

    assert validator
    assert list(validator(6, [])) == []
    assert list(validator(1, []))
    assert list(validator(-9, [])) == []
    assert list(validator(-7, []))

    defn = {"type": "number", "multipleOf": 3.14}
    validator = multiple_of(defn, mock.Mock())

    assert validator
    assert list(validator(6.28, [])) == []
    assert list(validator(1, []))
    assert list(validator(-9.42, [])) == []
    assert list(validator(-7, []))


def test_enum():
    defn = {"type": "string", "enum": list("ab")}
    validator = enum(defn, mock.Mock())

    assert list(validator("a", [])) == []
    assert list(validator("b", [])) == []
    assert list(validator("c", []))


def test_const():
    defn = {"type": "string", "const": "a"}
    validator = const(defn, mock.Mock())
    assert list(validator("a", [])) == []
    assert list(validator("b", []))

    defn = {"type": "integer", "const": 3}
    validator = const(defn, mock.Mock())
    assert list(validator(3, [])) == []
    assert list(validator(5, []))


def test_format():
    defn = {"type": "string", "format": "date"}
    validator = format_(defn, mock.Mock(formats=FORMATS))

    assert validator
    assert list(validator("2020-01-01", [])) == []
    assert list(validator("oops", []))

    # Unkown format ignored
    defn = {"type": "string", "format": "oops"}
    validator = format_(defn, mock.Mock(formats=FORMATS))
    assert validator is None


def test_format_overrides():
    defn = {
        "properties": {
            "squishy": {"type": "string", "format": "squishy"},
            "date": {"type": "string", "format": "date"},
            "email": {"type": "string", "format": "email"},
        }
    }

    def is_squishy(x: str) -> bool:
        return "squishy" in x

    validator = Validator(defn, formats={"date": is_squishy, "squishy": is_squishy})
    # new format is respected
    assert validator.is_valid({"squishy": "squishy!"})
    assert not validator.is_valid({"squishy": "2020-01-01"})
    # override is respected
    assert validator.is_valid({"date": "squishy!"})
    assert not validator.is_valid({"date": "2020-01-01"})
    # other formats are retained
    assert validator.is_valid({"email": "x@y.z"})
    assert not validator.is_valid({"email": "squishy!"})

    # override is not permanent
    validator = Validator(defn)
    assert validator.is_valid({"date": "2020-01-01"})


@pytest.mark.parametrize("wrap_testcase", (True, False))
@pytest.mark.parametrize(
    "wrapped,testcase,equal",
    (
        (True, True, True),
        (True, False, False),
        (True, 1, False),
        (True, 0, False),
        (False, False, True),
        (False, 0, False),
        (1, 1, True),
        (1, True, False),
        (1, 5, False),
        (0, 0, True),
        (0, False, False),
        (0, True, False),
        (0, 1, False),
    ),
)
def test_strict_bool(wrapped, testcase, equal, wrap_testcase):
    if wrap_testcase:
        testcase = _StrictBool(testcase)

    if equal:
        assert _StrictBool(wrapped) == testcase
        assert len({_StrictBool(wrapped), testcase}) == 1
    else:
        assert _StrictBool(wrapped) != testcase
        assert len({_StrictBool(wrapped), testcase}) == 2



from __future__ import annotations

import time

import pytest

from jsonscreamer import Validator

from .test_complex import POST_BODY, SCHEMA


@pytest.mark.parametrize(
    "variant",
    (
        "jsonschema",
        "jsonscreamer",
        "fastjsonschema",
    ),
)
def test_validate(variant: str):
    if variant == "jsonscreamer":
        validator = Validator(SCHEMA)

    elif variant == "jsonschema":
        from jsonschema import Draft7Validator

        Draft7Validator.check_schema(SCHEMA)

        validator = Draft7Validator(SCHEMA)
    else:
        import fastjsonschema

        class FastValidator:
            def __init__(self):
                self._validator = fastjsonschema.compile(SCHEMA)

            def is_valid(self, item):
                try:
                    self._validator(item)  # type: ignore
                    return True
                except Exception:
                    return False

        validator = FastValidator()

    t0 = time.monotonic()
    for _ in range(10_000):
        validator.is_valid(POST_BODY)
    print(time.monotonic() - t0)
